/*
 * PokeFrog - A 'Pokemonized' verison of Frogger, the well-known Konami game.
 *
 * copyright (c) 2012 Fabien LOISON <http://www.flogisoft.com/>
 * 
 * this program is free software. it comes without any warranty, to
 * the extent permitted by applicable law. you can redistribute it
 * and/or modify it under the terms of the do what the fuck you want
 * to public license, version 2, as published by sam hocevar. see
 * http://sam.zoy.org/wtfpl/copying for more details. 
 */


/**
 * @module PokeFrog
 */


/**
 * Manages all the game.
 *
 * @class Game
 */
function Game()
{
	// Tilesets
	this._tsMain = new wjs.Tileset(
			wjs.loader.getImage('main'),
			RES_TILESET_MAIN);
	this._tsPokemon = new wjs.Tileset(
			wjs.loader.getImage('pokemon'),
			RES_TILESET_POKEMON);
	this._tsEnemies = new wjs.Tileset(
			wjs.loader.getImage('enemies'),
			RES_TILESET_ENEMIES);
	this._tsFont = new wjs.Tileset(
			wjs.loader.getImage('font'),
			RES_TILESET_FONT);

	// Sounds
	this._sndDie = wjs.loader.getSound('die');
	this._sndLily = wjs.loader.getSound('lily');
	this._sndTime = wjs.loader.getSound('timealert');
	this._sndGameover = wjs.loader.getSound('gameover');
	this._sndWin = wjs.loader.getSound('win');

	// Background map
	this._mapBg = new wjs.Map(this._tsMain, 32, 32, RES_MAP_BG);

	// Non-walkable map
	this.cmap = [
		[0, 0, 32, 64],
		[96, 0, 160, 64],
		[224, 0, 288, 64],
		[352, 0, 416, 64],
		[480, 0, 544, 64],
		[608, 0, 650, 64]];

	// Lilies
	this._lilies = [
		new Lily(this._tsMain, 32, 0, 64, 64),
		new Lily(this._tsMain, 160, 0, 64, 64),
		new Lily(this._tsMain, 288, 0, 64, 64),
		new Lily(this._tsMain, 416, 0, 64, 64),
		new Lily(this._tsMain, 544, 0, 64, 64)];

	// Player
	this.player = null;
	this._resetPlayer();

	// Enemies
	this._enemies = [];
	for (var i=0 ; i<5 ; i++)
	{
		this._enemies.push({
			'orientation': (i%2) ? 'left' : 'right',
			'speed': Math.floor((Math.random()*3)+1),
			'next': 0,
			'y': 253 + 32 * i - i * 2,
			'enemies': []
		});
	}

	// Plateforms
	this._plateforms = [];
	for (var i=0 ; i<5 ; i++)
	{
		this._plateforms.push({
			'orientation': (i%2) ? 'left' : 'right',
			'speed': Math.floor((Math.random()*2)+1),
			'next': 0,
			'y': 64 + 32 * i,
			'plateforms': []
		});
	}

	// Lives
	this._lives = 3;

	// Time
	this._time = 60;
	this._mtime = 0;

	// Game finished (win or gameover)
	this._isFinished = false;
	this._finishScreenCb = null;

	// Game paused
	this.paused = false;

	// Register the refresh method on the main loop of wjs.
	wjs.mainLoop.addCallback('game', this.refresh.bind(this));
}

/**
 * End the game.
 *
 * @method end
 */
Game.prototype.end = function()
{
	wjs.mainLoop.delCallback('game');
	if (this.player)
	{
		this.player.die();
		this.player = null;
	}
	game = null;
}

/**
 * The player die (live-1,...)
 *
 * @method playerDie
 */
Game.prototype.playerDie = function()
{
	this._sndDie.play();
	this.player.die();
	this.paused = true;
	this._lives -= 1;
	if (this._lives <= 0)
	{
		this._lives = 0;
	}
	setTimeout(function() {
		game.paused = false;
		game._resetPlayer();
	}, 1000);
}

/**
 * Refresh the game state and the canvas.
 *
 * @method refresh
 */
Game.prototype.refresh = function()
{
	if (!this._isFinished && !this.paused)
	{
		this.update();
	}
	this.draw();
	if (this._isFinished && this._finishScreenCb)
	{
		this._finishScreenCb();
	}
}

/**
 * Update the game state (player'move, enemies, game over,...)
 *
 * @method update
 */
Game.prototype.update = function()
{
	// Update player
	game.player.update();

	// Update enemies
	for (var line in this._enemies)
	{
		// New enemy
		this._enemies[line].next -= 1;
		if (this._enemies[line].next <= 0 && wjs.helpers.arrayLength(this._enemies[line].enemies) < 4)
		{
			this._enemies[line].next = Math.floor((Math.random()*180)+80);
			this._enemies[line].enemies.push(new Enemy(
						this._tsEnemies,
						'truck-' + this._enemies[line].orientation,
						this._enemies[line].orientation,
						this._enemies[line].speed,
						this._enemies[line].y,
						[0, 10, 64, 32]
						));	
		}
		// Update enemies
		for (var i in this._enemies[line].enemies)
		{
			this._enemies[line].enemies[i].update();
			// Check for collide with player
			if (wjs.collide.checkRectRect(this.player.getHitbox(), this._enemies[line].enemies[i].getHitbox()))
			{
				this.playerDie();
			}
			// Delete enemies that are out of the screen.
			if (this._enemies[line].enemies[i].outOfScreen)
			{
				delete this._enemies[line].enemies[i];
			}
		}
	}

	// Update plateforms
	var platformCollid = false;
	for (var line in this._plateforms)
	{
		// New plateform
		this._plateforms[line].next -= 1;
		if (this._plateforms[line].next <= 0 && wjs.helpers.arrayLength(this._plateforms[line].plateforms) < 2)
		{
			this._plateforms[line].next = Math.floor((Math.random()*300)+100);
			this._plateforms[line].plateforms.push(new Enemy(
						this._tsMain,
						'trunk',
						this._plateforms[line].orientation,
						this._plateforms[line].speed,
						this._plateforms[line].y,
						[10, 2, 86, 21]
						));	
		}
		// Update palteforms
		for (var i in this._plateforms[line].plateforms)
		{
			this._plateforms[line].plateforms[i].update();
			// Check for collide with player
			if (!platformCollid && wjs.collide.checkRectRect(this.player.getHitbox(), this._plateforms[line].plateforms[i].getHitbox()))
			{
				platformCollid = true;
				if (this._plateforms[line].orientation == 'left')
				{
					this.player._x -= this._plateforms[line].speed;
				}
				else
				{
					this.player._x += this._plateforms[line].speed;
				}
			}
			// Delete plateforms that are out of the screen.
			if (this._plateforms[line].plateforms[i].outOfScreen)
			{
				delete this._plateforms[line].plateforms[i];
			}
		}
	}

	// Update time
	this._mtime -= 1;
	if (this._mtime <= 0)
	{
		this._mtime = 60;
		this._time -= 1;
		if (this._time < 0)
		{
			this._time = 0;
			this.playerDie();
		}
		if (this._time == 10 || this._time == 5)
		{
			this._sndTime.play();
		}
	}

	// Check "gloup gloup" :p
	if (!platformCollid && this.player.getHitbox()[3] > 65 && this.player.getHitbox()[3] < 224)
	{
		this.playerDie();
	}

	// Check collid with lilies
	this._checkCollideLilies();

	// Check end of game
	this._checkGameOver();
	this._checkWin();
}

/**
 * Draw all the game (background, player, enemies,...) on the canvas.
 *
 * @method draw
 */
Game.prototype.draw = function()
{
	// Draw map bg
	this._mapBg.draw();

	// Draw lilies
	for (var i in this._lilies)
	{
		this._lilies[i].draw();
	}

	// Draw enemies
	for (var line in this._enemies)
	{
		for (var i in this._enemies[line].enemies)
		{
			this._enemies[line].enemies[i].draw();
		}
	}

	// Draw plateforms
	for (var line in this._plateforms)
	{
		for (var i in this._plateforms[line].plateforms)
		{
			this._plateforms[line].plateforms[i].draw();
		}
	}

	// Draw player
	if (this.player)
	{
		this.player.draw();
	}

	// Draw lives
	this._tsMain.blit('darkbg0', 0, 420);
	this._tsMain.blit('lives'+this._lives, 0, 420);

	// Draw Time
	this._tsMain.blit('darkbg0', -36, 394);
	var fontColor = (this._time > 10) ? 'w' : 'r';
	this._tsFont.blit(
			fontColor + Math.floor(this._time / 10),
			2, 402);
	this._tsFont.blit(
			fontColor + this._time % 10,
			18, 402);

	// Update tilesets
	this._tsMain.update();
	this._tsPokemon.update();
	this._tsEnemies.update();

	// DEBUG
	if (DEBUG)
	{
		var ctx = wjs.canvas.getContext('2d');
		ctx.lineWidth = 2;
		ctx.strokeStyle = '#0f0';
		ctx.strokeRect(
				0,   // x
				64,  // y
				wjs.canvas.width,  // width
				224-64);  // height
		ctx.strokeStyle = '#00f';
		for (var czone in this.cmap)
		{
			ctx.strokeRect(
				this.cmap[czone][0],   // x
				this.cmap[czone][1],   // y
				this.cmap[czone][2] - this.cmap[czone][0],   // width
				this.cmap[czone][3] - this.cmap[czone][1]);  // height
		}
	}
}

/**
 * Check if the player lost.
 *
 * @method _checkGameOver
 */
Game.prototype._checkGameOver = function()
{
	if (this._lives <= 0 && !this.paused)
	{
		this._sndGameover.play();
		this._isFinished = true;
		this.player.die();
		
		var fs = new FinishScreen(
				wjs.loader.getImage('gameoverscreen'),
				function() {
					game.end();
					game = new Game();
				});
		this._finishScreenCb = fs.draw.bind(fs);
	}
}

/**
 * Check if the player won.
 *
 * @method _checkWin
 */
Game.prototype._checkWin = function()
{
	var count = 0;

	for (var i in this._lilies)
	{
		if (this._lilies[i].taken)
		{
			count += 1;
		}
		else
		{
			return;
		}
	}

	if (count == this._lilies.length && !this.paused)
	{
		setTimeout(function() {
			this._sndWin.play();
		}.bind(this), 1000);
		this._isFinished = true;
		this.player.die();
		
		var fs = new FinishScreen(
				wjs.loader.getImage('winscreen'),
				function() {
					game.end();
					game = new Game();
				});
		this._finishScreenCb = fs.draw.bind(fs);
	}
}

/**
 * Check collides with lilies and update the games status.
 *
 * @method _checkCollideLilies
 * @private
 */
Game.prototype._checkCollideLilies = function()
{
	for (var i in this._lilies)
	{
		if (wjs.collide.checkRectRect(this.player.getHitbox(), this._lilies[i].getHitbox()))
		{
			if (this._lilies[i].taken)
			{
				this.playerDie();
			}
			else
			{
				this._sndLily.play();
				this.paused = true;
				this._lilies[i].taken = true;
				this.player._visible = false;
				setTimeout(function() {
					game.paused = false;
					game._resetPlayer();
				}, 1000);
			}
		}
	}
}

/**
 * Kill the current player if exists, and create a new one at the start point.
 *
 * @method _resetPlayer
 * @private
 */
Game.prototype._resetPlayer = function()
{
	if (this.player)
	{
		this.player.die();
	}
	this.player = new Player(this._tsPokemon, RES_HITBOX_POKEMON);
	this._time = 60;
	this._mtime = 0;
}
